setGeneric('draw_title', function(object, ...) standardGeneric('draw_title'))
setGeneric('draw_heatmap_body', function(object, ...) standardGeneric('draw_heatmap_body'))
setGeneric('annotation_legend_size', function(object, ...) standardGeneric('annotation_legend_size'))
setGeneric('draw_annotation_legend', function(object, ...) standardGeneric('draw_annotation_legend'))
setGeneric('attach_annotation', function(object, ...) standardGeneric('attach_annotation'))
setGeneric('get_legend_param_list', function(object, ...) standardGeneric('get_legend_param_list'))
setGeneric('row_dend', function(object, ...) standardGeneric('row_dend'))
setGeneric('row_order', function(object, ...) standardGeneric('row_order'))
setGeneric('draw_heatmap_list', function(object, ...) standardGeneric('draw_heatmap_list'))
setGeneric('set_component_height', function(object, ...) standardGeneric('set_component_height'))
setGeneric('draw', function(object, ...) standardGeneric('draw'))
setGeneric('draw_annotation', function(object, ...) standardGeneric('draw_annotation'))
setGeneric('add_heatmap', function(object, ...) standardGeneric('add_heatmap'))
setGeneric('adjust_heatmap_list', function(object, ...) standardGeneric('adjust_heatmap_list'))
setGeneric('draw_heatmap_legend', function(object, ...) standardGeneric('draw_heatmap_legend'))
setGeneric('make_row_cluster', function(object, ...) standardGeneric('make_row_cluster'))
setGeneric('color_mapping_legend', function(object, ...) standardGeneric('color_mapping_legend'))
setGeneric('component_width', function(object, ...) standardGeneric('component_width'))
setGeneric('map_to_colors', function(object, ...) standardGeneric('map_to_colors'))
setGeneric('set_component_width', function(object, ...) standardGeneric('set_component_width'))
setGeneric('re_size', function(object, ...) standardGeneric('re_size'))
setGeneric('copy_all', function(object, ...) standardGeneric('copy_all'))
setGeneric('draw_dend', function(object, ...) standardGeneric('draw_dend'))
setGeneric('draw_dimnames', function(object, ...) standardGeneric('draw_dimnames'))
setGeneric('column_order', function(object, ...) standardGeneric('column_order'))
setGeneric('prepare', function(object, ...) standardGeneric('prepare'))
setGeneric('make_layout', function(object, ...) standardGeneric('make_layout'))
setGeneric('column_dend', function(object, ...) standardGeneric('column_dend'))
setGeneric('make_column_cluster', function(object, ...) standardGeneric('make_column_cluster'))
setGeneric('get_color_mapping_list', function(object, ...) standardGeneric('get_color_mapping_list'))
setGeneric('component_height', function(object, ...) standardGeneric('component_height'))
setGeneric('heatmap_legend_size', function(object, ...) standardGeneric('heatmap_legend_size'))
